/* eslint-disable @typescript-eslint/class-name-casing */
/**
 *  Copyright (c) 2020, Texas Instruments Incorporated
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:\
 *
 *  *   Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  notice, this list of conditions and the following disclaimer in the
 *  documentation and/or other materials provided with the distribution.
 *  *   Neither the name of Texas Instruments Incorporated nor the names of
 *  its contributors may be used to endorse or promote products derived
 *  from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import { NamedDataRecord, NamedRecordFieldDescriptor } from './NamedDataRecord';

const XDS110_VID = 0x0451;
const XDS110_PID_X1 = 0xbef1; // unassigned
const XDS110_PID_X2 = 0xbef2; // unassigned
const XDS110_PID_LEGACY = 0xbef3; // XDS110 Legacy config (no Drag-n-Drop)
const XDS110_PID_DND = 0xbef4; // XDS110 Drag-n-Drop config (no ET)

const XDS110ST_VID = 0x1cbe;
const XDS110ST_PID_LEGACY = 0x029e; // XDS110ST Legacy config (no Drag-n-Drop)
const XDS110ST_PID_TRACE = 0x029f; // XDS110ST Serial Trace config (debug + trace only)

const XDS110_CMSIS2_VID  = 0x1cbe;
const XDS110_CMSIS2_PID = 0x02a5; // XDS110 with CMSIS-DAP 2.0, ET, Drag-n-Drop

const DEFAULT_ATTEMPTS = 1;
const DEFAULT_TIMEOUT = 4000; // 4 seconds.
const XDS_SETSRST = 0x0e;    // Set level of SRST

const OFFSET_NONE = 0;
const OFFSET_XDS110 = -260;
const SC_ERR_NONE = (OFFSET_NONE - 0);
const SC_ERR_XDS110_OPEN = (OFFSET_XDS110 - 0);
const SC_ERR_XDS110_FAIL = (OFFSET_XDS110 - 1);

enum DEBUGSTSTUS {
    DEBUG_OK = 0,       // !< Function was successful
    DEBUG_FAIL         // !< An error occurred
};

declare namespace gc {
    const services: { [index: string]: unknown };
};

function usbDeviceListReady(): Promise<Map<string, any>> {
    const usbService: any = gc.services['ti-service-usb'];
    if (usbService && !usbService.isUsbListReady) {
        return new Promise<Map<string, any>>( (resolve) => {
            usbService.addEventListener('usbDeviceListUpdated', () => {
                resolve(usbService.deviceInfoMap as Map<string, any>);
            });
        });
    }
    return Promise.resolve(usbService.deviceInfoMap);
};

function getUsbDeviceDescriptor(key: string): Promise<any> {
    const usbService: any = gc.services['ti-service-usb'];
    return new Promise<any>( (resolve, reject ) => {
        usbService.getDescriptors(key).then(resolve).fail(reject);
    });
};


function getSerialNumber(key: string): Promise<string> {
    // TODO: implement reading serial number
    return Promise.resolve('');
};

/**
 * Open the low level USB Interface
 */
function openUsbDevice(key: string, nInterface: number): Promise<void> {
    const usbService: any = gc.services['ti-service-usb'];
    return new Promise<void>( (resolve, reject) => {
        usbService.openUsbDevice(key).then(() => {
            return usbService.claimInterface(key, nInterface, true, 1);
        }).then(resolve).fail(reject);
    });
};

/**
 * Close the low level USB Interface
 */
function closeUsbDevice(key: string, nInterface: number): Promise<void> {
    const usbService: any = gc.services['ti-service-usb'];
    return new Promise<void>( (resolve, reject) => {
        usbService.releaseInterface(key, nInterface).then(() => {
            usbService.closeUsbDevice(key).then(resolve).fail(reject);
        });
    });
};

function sendCmd(key: string, nInterface: number, data: Uint8Array, timeout: number): Promise<Uint8Array> {
    const usbService: any = gc.services['ti-service-usb'];
    return new Promise<Uint8Array>( (resolve, reject) => {
        // convert to hex string for transmission
        const cmd = data.reduce( (result: string, byte, index) => result += (index === 0 ? '0x' : ',0x') + byte.toString(16), '');
        usbService.sendCmd(key, nInterface, cmd, timeout, null).then((rxData: any) => {
            resolve(rxData.data);
        }).fail(reject);
    });
};

/**
 *
 * Open connection to the XDS110.
 *
 * This function will connect to the XDS110, but not request the debug probe to
 * perform ANY initialization in preparation for JTAG/SWD scanning operations.
 *
 */
async function XDS_Open(serial?: string): Promise<[string, number]> {
    const vids = [ XDS110_VID,        XDS110_VID,     XDS110ST_VID,        XDS110ST_VID,       XDS110_CMSIS2_VID ];
    const pids = [ XDS110_PID_LEGACY, XDS110_PID_DND, XDS110ST_PID_LEGACY, XDS110ST_PID_TRACE, XDS110_CMSIS2_PID ];

    const usbDeviceInfoMap: Map<string, any> = await usbDeviceListReady();
    let key: string | undefined = undefined;
    let debugInterface: number | undefined;

    for (key of usbDeviceInfoMap.keys()) {
        const deviceInfo = await getUsbDeviceDescriptor(key);

        for (let i = 0; i < pids.length; i++) {
            if (deviceInfo.deviceDescriptor.idProduct === pids[i] && deviceInfo.deviceDescriptor.idVendor === vids[i]) {
                if (!serial || serial === await getSerialNumber(key)) {
                    // Get the interface and endpoint data for device
                    debugInterface = 2;
                    if ((XDS110ST_VID === vids[i] && XDS110ST_PID_TRACE === pids[i]) ||
                        (XDS110_CMSIS2_VID === vids[i] && XDS110_CMSIS2_PID === pids[i])) {
                        // XDS110ST in Trace Performance configuration
                        // XDS110 CMSIS-DAP 2.0 configuration
                        debugInterface = 0;
                    }

                    await openUsbDevice(key, debugInterface);
                    return [key, debugInterface];
                }
            }
        }
    }
    throw SC_ERR_XDS110_OPEN;
};

/**
 *
 * Close connection to the XDS110.
 *
 */
function XDS_Close(key: string, debugInterface: number): Promise<void> {
    if (key && debugInterface !== undefined) {
        return closeUsbDevice(key, debugInterface);
    }
    return Promise.resolve();
};

interface IXDS_Out_Header {
    startByte: number;
    size: number;
    cmd: number;
};

interface IXDS_Out_SetSrts extends IXDS_Out_Header{
    data: number;
};

class XDS_Out_Header extends NamedDataRecord<IXDS_Out_Header> {
    littleEndian = true;
    static fieldDescriptors: NamedRecordFieldDescriptor<IXDS_Out_Header>[] = [
        ['startByte', 1],
        ['size', 2],
        ['cmd', 1]
    ];
};

class XDS_Out_SetSrst extends NamedDataRecord<IXDS_Out_SetSrts> {
    littleEndian = true;
    static fieldDescriptors = (XDS_Out_Header.fieldDescriptors as NamedRecordFieldDescriptor<IXDS_Out_SetSrts>[]).concat([
        ['data', 1]
    ]);
};

interface IXDS_In_Header {
    startByte: number;
    size: number;
    error: number;
};

class XDS_In_Header extends NamedDataRecord<IXDS_In_Header> {
    littleEndian = true;
    static fieldDescriptors: NamedRecordFieldDescriptor<IXDS_In_Header>[] = [
        ['startByte', 1],
        ['size', 2],
        ['error', 4]
    ];
};

/**
 * Execute XDS command packet via USB packet handling.
 *
 * This function will send an XDS command packet to the debug probe.
 *
 * @param key
 * @param debugInterface
 * @param packet
 * @param responseLength
 * @param attempts
 * @param timeout
 *
 * @returns JTI_Error code
 */
async function XDS_Execute(key: string, debugInterface: number, packet: IXDS_Out_Header & NamedDataRecord<IXDS_Out_Header>,
    responseLength: number, attempts = DEFAULT_ATTEMPTS, timeout = DEFAULT_TIMEOUT) {
    const START_BYTE = 42;  // ascii *
    packet.startByte = START_BYTE;
    packet.size = packet.length - 1 - 2;  // size minus start byte and size;

    let status = DEBUGSTSTUS.DEBUG_FAIL;
    let done = false;
    let result: Uint8Array | undefined;
    while (!done && attempts > 0) {
        attempts--;
        try {
            result = await sendCmd(key, debugInterface, packet.asUint8Array, timeout);
            if (result.length === responseLength) {
                status = DEBUGSTSTUS.DEBUG_OK;
                done = true;
            }
        // eslint-disable-next-line no-empty
        } catch {
        }
    }

    // If packet was received OK, then do sanity tests on content.
    // Check start character
    if (status === DEBUGSTSTUS.DEBUG_OK && result) {
        const header = XDS_In_Header.create(result.slice(0, XDS_In_Header.getSize()));
        if (header.startByte !== START_BYTE || (header.size !== responseLength - 1 -2)) {
            status = DEBUGSTSTUS.DEBUG_FAIL;
        } else if (header.error !== SC_ERR_NONE) {
            throw header.error;
        }
    } else {
        throw SC_ERR_XDS110_FAIL;
    }

    return result;
}

async function XDS_SetSrst(key: string, debugInterface: number, srst: number) {

    const packet = XDS_Out_SetSrst.create();
    packet.cmd = XDS_SETSRST;
    packet.data = srst;

    await XDS_Execute(key, debugInterface, packet, XDS_In_Header.getSize());
};

export async function xds110reset(assert: boolean, log?: (message: string) => void, serial?: string ) {
    log = log || (() => {});

    try {
        const [key, debugInterface] = await XDS_Open(serial);
        try {
            await XDS_SetSrst(key, debugInterface, assert ? 0 : 1);
        } catch (error) {
            log(`ERROR: Failed to ${assert ? 'assert' : 'deassert'} nSRST ${error}.`);
        } finally {
            XDS_Close(key, debugInterface);
        }
    } catch (error) {
        log(`ERROR: Failed to connect to XDS110 ${error}. Check USB connection ${serial ? 'and/or serial number.' : 'to the XDS110.'}`);
    }
};